/*
 * AutomatonToStringComposer.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef AUTOMATON_TO_STRING_COMPOSER_H_
#define AUTOMATON_TO_STRING_COMPOSER_H_

#include <ostream>
#include <range>
#include <core/multipleDispatch.hpp>
#include "automaton/Automaton.h"
#include "automaton/AutomatonFeatures.h"
#include "label/LabelFeatures.h"

#include <object/ObjectToStringComposer.h>

namespace automaton {

class AutomatonToStringComposer : public alib::SingleDispatchFirstStaticParam<AutomatonToStringComposer, void, std::ostream&, const AutomatonBase &> {
	template < class SymbolType, class StateType >
	static void composeTransitionsFromState(std::ostream& out, const DFA < SymbolType, StateType > & automaton, const StateType & from);
	template < class SymbolType, class StateType >
	static void composeTransitionsFromState(std::ostream& out, const NFA < SymbolType, StateType > & automaton, const StateType & from);
	template < class SymbolType, class StateType >
	static void composeTransitionsFromState(std::ostream& out, const MultiInitialStateNFA < SymbolType, StateType > & automaton, const StateType & from);
	template < class SymbolType, class EpsilonType, class StateType >
	static void composeTransitionsFromState(std::ostream& out, const EpsilonNFA < SymbolType, EpsilonType, StateType > & automaton, const StateType & from);

public:
	/**
	 * Prints XML representation of UnknownAutomaton to the output stream.
	 * @param automaton automaton to print
	 * @return list of xml tokens representing the automaton
	 */
	static void compose(std::ostream& output, const Automaton& automaton);

	template < class SymbolType, class StateType >
	static void compose(std::ostream& output, const DFA < SymbolType, StateType > & automaton);
	template < class SymbolType, class StateType >
	static std::string compose(const DFA < SymbolType, StateType > & automaton);
	template < class SymbolType, class StateType >
	static void compose(std::ostream& output, const NFA < SymbolType, StateType > & automaton);
	template < class SymbolType, class StateType >
	static void compose(std::ostream& output, const MultiInitialStateNFA < SymbolType, StateType > & automaton);
	template < class SymbolType, class EpsilonType, class StateType >
	static void compose(std::ostream& output, const EpsilonNFA < SymbolType, EpsilonType, StateType > & automaton);
};

template < class SymbolType, class StateType >
void AutomatonToStringComposer::composeTransitionsFromState(std::ostream& out, const DFA < SymbolType, StateType > & automaton, const StateType & from) {
	ext::range < typename std::map < std::pair < StateType, SymbolType >, StateType >::const_iterator > symbolTransitionsFromState = automaton.getTransitionsFromState(from);

	for(const SymbolType& inputSymbol : automaton.getInputAlphabet()) {
		if ( ! symbolTransitionsFromState.empty ( ) && symbolTransitionsFromState.front ( ).first.second == inputSymbol ) {
			out << " ";
			alib::ObjectToStringComposer::compose(out, symbolTransitionsFromState.front ( ).second);
			symbolTransitionsFromState.pop_front ( );
		} else {
			out << " -";
		}
	}
}

template < class SymbolType, class StateType >
void AutomatonToStringComposer::composeTransitionsFromState(std::ostream& out, const NFA < SymbolType, StateType > & automaton, const StateType & from) {
	std::map<std::pair<StateType, SymbolType>, std::set<StateType> > symbolTransitionsFromState = automaton.getTransitionsFromState(from);

	for(const SymbolType& inputSymbol : automaton.getInputAlphabet()) {
		const typename std::map<std::pair<StateType, SymbolType>, std::set<StateType> >::iterator toStates = symbolTransitionsFromState.find(std::make_pair(from, inputSymbol));
		if(toStates == symbolTransitionsFromState.end() || toStates->second.size() == 0) {
			out << " -";
		} else {
			bool sign = false;
			for(const StateType& to : toStates->second) {
				out << (sign ? "|" : " ");
				alib::ObjectToStringComposer::compose(out, to);
				sign = true;
			}
		}
	}
}

template < class SymbolType, class StateType >
void AutomatonToStringComposer::composeTransitionsFromState(std::ostream& out, const MultiInitialStateNFA < SymbolType, StateType > & automaton, const StateType & from) {
	std::map<std::pair<StateType, SymbolType>, std::set<StateType> > symbolTransitionsFromState = automaton.getTransitionsFromState(from);

	for(const SymbolType& inputSymbol : automaton.getInputAlphabet()) {
		const typename std::map<std::pair<StateType, SymbolType>, std::set<StateType> >::iterator toStates = symbolTransitionsFromState.find(std::make_pair(from, inputSymbol));
		if(toStates == symbolTransitionsFromState.end() || toStates->second.size() == 0) {
			out << " -";
		} else {
			bool sign = false;
			for(const StateType& to : toStates->second) {
				out << (sign ? "|" : " ");
				alib::ObjectToStringComposer::compose(out, to);
				sign = true;
			}
		}
	}
}

template < class SymbolType, class EpsilonType, class StateType >
void AutomatonToStringComposer::composeTransitionsFromState(std::ostream& out, const EpsilonNFA < SymbolType, EpsilonType, StateType > & automaton, const StateType & from) {
	std::map<std::pair<StateType, SymbolType>, std::set<StateType> > symbolTransitionsFromState = automaton.getSymbolTransitionsFromState(from);

	for(const SymbolType& inputSymbol : automaton.getInputAlphabet()) {
		const typename std::map<std::pair<StateType, SymbolType>, std::set<StateType> >::iterator toStates = symbolTransitionsFromState.find(std::make_pair(from, inputSymbol));
		if(toStates == symbolTransitionsFromState.end() || toStates->second.size() == 0) {
			out << " -";
		} else {
			bool sign = false;
			for(const StateType& to : toStates->second) {
				out << (sign ? "|" : " ");
				alib::ObjectToStringComposer::compose(out, to);
				sign = true;
			}
		}
	}

	std::map<StateType, std::set<StateType> > epsilonTransitionsFromState = automaton.getEpsilonTransitionsFromState(from);
	if(epsilonTransitionsFromState.size() == 0 || epsilonTransitionsFromState.begin()->second.size() == 0) {
		out << " -";
	} else {
		bool sign = false;
		for(const StateType& to : epsilonTransitionsFromState.begin()->second) {
			out << (sign ? "|" : " ");
			alib::ObjectToStringComposer::compose(out, to);
			sign = true;
		}
	}
}

} /* namespace automaton */

#endif /* AUTOMATON_TO_STRING_COMPOSER_H_ */
