/*
 * GrammarFromStringParser.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#include "GrammarFromStringParser.h"

#include "exception/CommonException.h"

#include "grammar/GrammarClasses.h"
#include <grammar/Grammar.h>

#include "../StringApi.hpp"

namespace grammar {

Grammar GrammarFromStringParser::parseGrammar(std::istream& input) {
	return parseGrammar(input, std::set<FEATURES>({FEATURES::LEFT_LG, FEATURES::LEFT_RG, FEATURES::RIGHT_LG, FEATURES::RIGHT_RG, FEATURES::LG, FEATURES::CFG, FEATURES::EPSILON_FREE_CFG, FEATURES::CNF, FEATURES::GNF, FEATURES::CSG, FEATURES::NON_CONTRACTING_GRAMMAR, FEATURES::CONTEXT_PRESERVING_UNRESTRICTED_GRAMMAR, FEATURES::UNRESTRICTED_GRAMMAR}));
}

Grammar GrammarFromStringParser::parseGrammar(std::istream& input, const std::set<FEATURES>& features) {
	GrammarFromStringLexer::Token token = GrammarFromStringLexer::next(input);
	if(token.type == GrammarFromStringLexer::TokenType::LEFT_RG) {
		if(!features.count(FEATURES::LEFT_RG)) throw exception::CommonException("Disabled Formalism LeftRG");
		GrammarFromStringLexer::putback(input, token);
		return Grammar(parseLeftRG(input));
	} else if(token.type == GrammarFromStringLexer::TokenType::RIGHT_RG) {
		if(!features.count(FEATURES::RIGHT_RG)) throw exception::CommonException("Disabled Formalism RightRG");
		GrammarFromStringLexer::putback(input, token);
		return Grammar(parseRightRG(input));
	} else if(token.type == GrammarFromStringLexer::TokenType::LEFT_LG) {
		if(!features.count(FEATURES::LEFT_LG)) throw exception::CommonException("Disabled Formalism LeftLG");
		GrammarFromStringLexer::putback(input, token);
		return Grammar(parseLeftLG(input));
	} else if(token.type == GrammarFromStringLexer::TokenType::RIGHT_LG) {
		if(!features.count(FEATURES::RIGHT_LG)) throw exception::CommonException("Disabled Formalism RightLG");
		GrammarFromStringLexer::putback(input, token);
		return Grammar(parseRightLG(input));
	} else if(token.type == GrammarFromStringLexer::TokenType::LG) {
		if(!features.count(FEATURES::LG)) throw exception::CommonException("Disabled formalism LG");
		GrammarFromStringLexer::putback(input, token);
		return Grammar(parseLG(input));
	} else if(token.type == GrammarFromStringLexer::TokenType::CFG) {
		if(!features.count(FEATURES::CFG)) throw exception::CommonException("Disabled formalism CFG");
		GrammarFromStringLexer::putback(input, token);
		return Grammar(parseCFG(input));
	} else if(token.type == GrammarFromStringLexer::TokenType::EPSILON_FREE_CFG) {
		if(!features.count(FEATURES::EPSILON_FREE_CFG)) throw exception::CommonException("Disabled formalism EpsilonFreeCFG");
		GrammarFromStringLexer::putback(input, token);
		return Grammar(parseEpsilonFreeCFG(input));
	} else if(token.type == GrammarFromStringLexer::TokenType::GNF) {
		if(!features.count(FEATURES::GNF)) throw exception::CommonException("Disabled formalism GNF");
		GrammarFromStringLexer::putback(input, token);
		return Grammar(parseGNF(input));
	} else if(token.type == GrammarFromStringLexer::TokenType::CNF) {
		if(!features.count(FEATURES::CNF)) throw exception::CommonException("Disabled formalism CNF");
		GrammarFromStringLexer::putback(input, token);
		return Grammar(parseCNF(input));
	} else if(token.type == GrammarFromStringLexer::TokenType::NON_CONTRACTING_GRAMMAR) {
		if(!features.count(FEATURES::NON_CONTRACTING_GRAMMAR)) throw exception::CommonException("Disabled formalism NonContractingGrammar");
		GrammarFromStringLexer::putback(input, token);
		return Grammar(parseNonContractingGrammar(input));
	} else if(token.type == GrammarFromStringLexer::TokenType::CSG) {
		if(!features.count(FEATURES::CSG)) throw exception::CommonException("Disabled formalism CSG");
		GrammarFromStringLexer::putback(input, token);
		return Grammar(parseCSG(input));
	} else if(token.type == GrammarFromStringLexer::TokenType::CONTEXT_PRESERVING_UNRESTRICTED_GRAMMAR) {
		if(!features.count(FEATURES::CONTEXT_PRESERVING_UNRESTRICTED_GRAMMAR)) throw exception::CommonException("Disabled formalism ContextPreservingUnrestrictedGrammar");
		GrammarFromStringLexer::putback(input, token);
		return Grammar(parseContextPreservingUnrestrictedGrammar(input));
	} else if(token.type == GrammarFromStringLexer::TokenType::UNRESTRICTED_GRAMMAR) {
		if(!features.count(FEATURES::UNRESTRICTED_GRAMMAR)) throw exception::CommonException("Disabled formalism UnrestrictedGrammar");
		GrammarFromStringLexer::putback(input, token);
		return Grammar(parseUnrestrictedGrammar(input));
	} else {
		throw exception::CommonException("Formalism not recognised (token = \"" + token.value + "\")");
	}
}

std::set<DefaultSymbolType> GrammarFromStringParser::parseSet(std::istream& input) {
	std::set<DefaultSymbolType> res;

	GrammarFromStringLexer::Token token = GrammarFromStringLexer::next(input);
	if(token.type != GrammarFromStringLexer::TokenType::SET_BEGIN) {
		throw exception::CommonException("Expected SET_BEGIN token.");
	}

	token = GrammarFromStringLexer::next(input);
	if(token.type != GrammarFromStringLexer::TokenType::SET_END) {
		GrammarFromStringLexer::putback(input, token);
		while(true) {
			DefaultSymbolType symbol = alib::stringApi<DefaultSymbolType>::parse(input);
			res.insert(symbol);

			token = GrammarFromStringLexer::next(input);
			if(token.type == GrammarFromStringLexer::TokenType::SET_END) {
				break;
			}
			if(token.type != GrammarFromStringLexer::TokenType::COMMA) {
				throw exception::CommonException("Expected SET_END or COMMA token");
			}
		}
	}

	if(token.type != GrammarFromStringLexer::TokenType::SET_END) {
		throw exception::CommonException("Expected SET_END token");
	}

	return res;
}

std::map<DefaultSymbolType, std::set<std::vector<DefaultSymbolType>>> GrammarFromStringParser::parseCFLikeRules(std::istream& input) {
	std::map<DefaultSymbolType, std::set<std::vector<DefaultSymbolType>>> result;

	GrammarFromStringLexer::Token token = GrammarFromStringLexer::next(input);
	if(token.type != GrammarFromStringLexer::TokenType::SET_BEGIN) {
		throw exception::CommonException("Expected SET_BEGIN token.");
	}

	token = GrammarFromStringLexer::next(input);
	if(token.type != GrammarFromStringLexer::TokenType::SET_END) {
		GrammarFromStringLexer::putback(input, token);
		while(true) {
			DefaultSymbolType lhs = alib::stringApi<DefaultSymbolType>::parse(input);

			token = GrammarFromStringLexer::next(input);
			if(token.type != GrammarFromStringLexer::TokenType::MAPS_TO) {
				throw exception::CommonException("Expected MAPS_TO token.");
			}

			while(true) {
				std::vector<DefaultSymbolType> rhs;

				token = GrammarFromStringLexer::next(input);
				if(token.type != GrammarFromStringLexer::TokenType::COMMA && token.type != GrammarFromStringLexer::TokenType::SET_END && token.type != GrammarFromStringLexer::TokenType::SEPARATOR) while(true) {
					GrammarFromStringLexer::putback(input, token);

					rhs.push_back(alib::stringApi<DefaultSymbolType>::parse(input));
					token = GrammarFromStringLexer::next(input);
					if(token.type == GrammarFromStringLexer::TokenType::SEPARATOR || token.type == GrammarFromStringLexer::TokenType::COMMA || token.type == GrammarFromStringLexer::TokenType::SET_END) {
						break;
					}
				}
				result[lhs].insert(rhs);
				if(token.type == GrammarFromStringLexer::TokenType::COMMA || token.type == GrammarFromStringLexer::TokenType::SET_END) {
					break;
				}
				if(token.type != GrammarFromStringLexer::TokenType::SEPARATOR) {
					throw exception::CommonException("Expected SEPARATOR, SETEND or COMMA token");
				}
			}

			if(token.type == GrammarFromStringLexer::TokenType::SET_END) {
				break;
			}
			if(token.type != GrammarFromStringLexer::TokenType::COMMA) {
				throw exception::CommonException("Expected SET_END or COMMA token");
			}
		}
	}

	if(token.type != GrammarFromStringLexer::TokenType::SET_END) {
		throw exception::CommonException("Expected SET_END token");
	}
	return result;
}

template<class T>
T GrammarFromStringParser::parseCFLikeGrammar(std::istream& input) {
	GrammarFromStringLexer::Token token = GrammarFromStringLexer::next(input);
	if(token.type != GrammarFromStringLexer::TokenType::TUPLE_BEGIN) {
		throw exception::CommonException("Unrecognised Tuple begin token.");
	}

	std::set<DefaultSymbolType> nonterminals = parseSet(input);

	token = GrammarFromStringLexer::next(input);
	if(token.type != GrammarFromStringLexer::TokenType::COMMA) {
		throw exception::CommonException("Unrecognised Comma token.");
	}

	std::set<DefaultSymbolType> terminals = parseSet(input);

	token = GrammarFromStringLexer::next(input);
	if(token.type != GrammarFromStringLexer::TokenType::COMMA) {
		throw exception::CommonException("Unrecognised Comma token.");
	}

	std::map<DefaultSymbolType, std::set<std::vector<DefaultSymbolType>>> rules = parseCFLikeRules(input);

	token = GrammarFromStringLexer::next(input);
	if(token.type != GrammarFromStringLexer::TokenType::COMMA) {
		throw exception::CommonException("Unrecognised Comma token.");
	}

	DefaultSymbolType initialSymbol = alib::stringApi<DefaultSymbolType>::parse(input);

	token = GrammarFromStringLexer::next(input);
	if(token.type != GrammarFromStringLexer::TokenType::TUPLE_END) {
		throw exception::CommonException("Unrecognised Tuple end token.");
	}

	T grammar(nonterminals, terminals, initialSymbol);
	for(const auto& rule : rules) {
		for(const auto& ruleRHS : rule.second) {
			grammar.addRawRule(rule.first, ruleRHS);
		}
	}
	return grammar;
}

std::map<std::vector<DefaultSymbolType>, std::set<std::vector<DefaultSymbolType>>> GrammarFromStringParser::parseCSLikeRules(std::istream& input) {
	std::map<std::vector<DefaultSymbolType>, std::set<std::vector<DefaultSymbolType>>> result;

	GrammarFromStringLexer::Token token = GrammarFromStringLexer::next(input);
	if(token.type != GrammarFromStringLexer::TokenType::SET_BEGIN) {
		throw exception::CommonException("Expected SET_BEGIN token.");
	}

	token = GrammarFromStringLexer::next(input);
	if(token.type != GrammarFromStringLexer::TokenType::SET_END) {
		GrammarFromStringLexer::putback(input, token);
		while(true) {
			std::vector<DefaultSymbolType> lhs;
			while(true) {
				lhs.push_back(alib::stringApi<DefaultSymbolType>::parse(input));
				token = GrammarFromStringLexer::next(input);
				if(token.type == GrammarFromStringLexer::TokenType::MAPS_TO) {
					break;
				}
				GrammarFromStringLexer::putback(input, token);
			}

			if(token.type != GrammarFromStringLexer::TokenType::MAPS_TO) {
				throw exception::CommonException("Expected MAPS_TO token.");
			}

			while(true) {
				std::vector<DefaultSymbolType> rhs;

				token = GrammarFromStringLexer::next(input);
				if(token.type != GrammarFromStringLexer::TokenType::COMMA && token.type != GrammarFromStringLexer::TokenType::SET_END && token.type != GrammarFromStringLexer::TokenType::SEPARATOR) while(true) {
					GrammarFromStringLexer::putback(input, token);

					rhs.push_back(alib::stringApi<DefaultSymbolType>::parse(input));
					token = GrammarFromStringLexer::next(input);
					if(token.type == GrammarFromStringLexer::TokenType::SEPARATOR || token.type == GrammarFromStringLexer::TokenType::COMMA || token.type == GrammarFromStringLexer::TokenType::SET_END) {
						break;
					}
				}
				result[lhs].insert(rhs);
				if(token.type == GrammarFromStringLexer::TokenType::COMMA || token.type == GrammarFromStringLexer::TokenType::SET_END) {
					break;
				}
				if(token.type != GrammarFromStringLexer::TokenType::SEPARATOR) {
					throw exception::CommonException("Expected SEPARATOR, SETEND or COMMA token");
				}
			}

			if(token.type == GrammarFromStringLexer::TokenType::SET_END) {
				break;
			}
			if(token.type != GrammarFromStringLexer::TokenType::COMMA) {
				throw exception::CommonException("Expected SET_END or COMMA token");
			}
		}
	}

	if(token.type != GrammarFromStringLexer::TokenType::SET_END) {
		throw exception::CommonException("Expected SET_END token");
	}
	return result;
}

template<class T>
T GrammarFromStringParser::parseCSLikeGrammar(std::istream& input) {
	GrammarFromStringLexer::Token token = GrammarFromStringLexer::next(input);
	if(token.type != GrammarFromStringLexer::TokenType::TUPLE_BEGIN) {
		throw exception::CommonException("Unrecognised Tuple begin token.");
	}

	std::set<DefaultSymbolType> nonterminals = parseSet(input);

	token = GrammarFromStringLexer::next(input);
	if(token.type != GrammarFromStringLexer::TokenType::COMMA) {
		throw exception::CommonException("Unrecognised Comma token.");
	}

	std::set<DefaultSymbolType> terminals = parseSet(input);

	token = GrammarFromStringLexer::next(input);
	if(token.type != GrammarFromStringLexer::TokenType::COMMA) {
		throw exception::CommonException("Unrecognised Comma token.");
	}

	std::map<std::vector<DefaultSymbolType>, std::set<std::vector<DefaultSymbolType>>> rules = parseCSLikeRules(input);

	token = GrammarFromStringLexer::next(input);
	if(token.type != GrammarFromStringLexer::TokenType::COMMA) {
		throw exception::CommonException("Unrecognised Comma token.");
	}

	DefaultSymbolType initialSymbol = alib::stringApi<DefaultSymbolType>::parse(input);

	token = GrammarFromStringLexer::next(input);
	if(token.type != GrammarFromStringLexer::TokenType::TUPLE_END) {
		throw exception::CommonException("Unrecognised Tuple end token.");
	}

	T grammar(nonterminals, terminals, initialSymbol);
	for(const auto& rule : rules) {
		for(const auto& ruleRHS : rule.second) {
			grammar.addRule(rule.first, ruleRHS);
		}
	}
	return grammar;
}

std::map<std::tuple<std::vector<DefaultSymbolType>, DefaultSymbolType, std::vector<DefaultSymbolType>>, std::set<std::vector<DefaultSymbolType>>> GrammarFromStringParser::parsePreservingCSLikeRules(std::istream& input) {
	std::map<std::tuple<std::vector<DefaultSymbolType>, DefaultSymbolType, std::vector<DefaultSymbolType>>, std::set<std::vector<DefaultSymbolType>>> result;

	GrammarFromStringLexer::Token token = GrammarFromStringLexer::next(input);
	if(token.type != GrammarFromStringLexer::TokenType::SET_BEGIN) {
		throw exception::CommonException("Expected SET_BEGIN token.");
	}

	token = GrammarFromStringLexer::next(input);
	if(token.type != GrammarFromStringLexer::TokenType::SET_END) {
		GrammarFromStringLexer::putback(input, token);
		while(true) {
			std::vector<DefaultSymbolType> lContext;
			token = GrammarFromStringLexer::next(input);
			if(token.type != GrammarFromStringLexer::TokenType::SEPARATOR) while(true) {
				GrammarFromStringLexer::putback(input, token);
				lContext.push_back(alib::stringApi<DefaultSymbolType>::parse(input));
				token = GrammarFromStringLexer::next(input);
				if(token.type == GrammarFromStringLexer::TokenType::SEPARATOR) {
					break;
				}
			}

			if(token.type != GrammarFromStringLexer::TokenType::SEPARATOR) {
				throw exception::CommonException("Expected SEPARATOR token.");
			}

			DefaultSymbolType lhs = alib::stringApi<DefaultSymbolType>::parse(input);

			token = GrammarFromStringLexer::next(input);
			if(token.type != GrammarFromStringLexer::TokenType::SEPARATOR) {
				throw exception::CommonException("Expected SEPARATOR token.");
			}

			std::vector<DefaultSymbolType> rContext;
			token = GrammarFromStringLexer::next(input);
			if(token.type != GrammarFromStringLexer::TokenType::MAPS_TO) while(true) {
				GrammarFromStringLexer::putback(input, token);
				rContext.push_back(alib::stringApi<DefaultSymbolType>::parse(input));
				token = GrammarFromStringLexer::next(input);
				if(token.type == GrammarFromStringLexer::TokenType::MAPS_TO) {
					break;
				}
			}

			if(token.type != GrammarFromStringLexer::TokenType::MAPS_TO) {
				throw exception::CommonException("Expected MAPS_TO token.");
			}

			std::tuple<std::vector<DefaultSymbolType>, DefaultSymbolType, std::vector<DefaultSymbolType>> key = std::make_tuple(lContext, lhs, rContext);

			while(true) {
				std::vector<DefaultSymbolType> rhs;

				token = GrammarFromStringLexer::next(input);
				if(token.type != GrammarFromStringLexer::TokenType::COMMA && token.type != GrammarFromStringLexer::TokenType::SET_END && token.type != GrammarFromStringLexer::TokenType::SEPARATOR) while(true) {
					GrammarFromStringLexer::putback(input, token);

					rhs.push_back(alib::stringApi<DefaultSymbolType>::parse(input));
					token = GrammarFromStringLexer::next(input);
					if(token.type == GrammarFromStringLexer::TokenType::SEPARATOR || token.type == GrammarFromStringLexer::TokenType::COMMA || token.type == GrammarFromStringLexer::TokenType::SET_END) {
						break;
					}
				}
				result[key].insert(rhs);
				if(token.type == GrammarFromStringLexer::TokenType::COMMA || token.type == GrammarFromStringLexer::TokenType::SET_END) {
					break;
				}
				if(token.type != GrammarFromStringLexer::TokenType::SEPARATOR) {
					throw exception::CommonException("Expected SEPARATOR, SETEND or COMMA token");
				}
			}

			if(token.type == GrammarFromStringLexer::TokenType::SET_END) {
				break;
			}
			if(token.type != GrammarFromStringLexer::TokenType::COMMA) {
				throw exception::CommonException("Expected SET_END or COMMA token");
			}
		}
	}

	if(token.type != GrammarFromStringLexer::TokenType::SET_END) {
		throw exception::CommonException("Expected SET_END token");
	}
	return result;
}

template<class T>
T GrammarFromStringParser::parsePreservingCSLikeGrammar(std::istream& input) {
	GrammarFromStringLexer::Token token = GrammarFromStringLexer::next(input);
	if(token.type != GrammarFromStringLexer::TokenType::TUPLE_BEGIN) {
		throw exception::CommonException("Unrecognised Tuple begin token.");
	}

	std::set<DefaultSymbolType> nonterminals = parseSet(input);

	token = GrammarFromStringLexer::next(input);
	if(token.type != GrammarFromStringLexer::TokenType::COMMA) {
		throw exception::CommonException("Unrecognised Comma token.");
	}

	std::set<DefaultSymbolType> terminals = parseSet(input);

	token = GrammarFromStringLexer::next(input);
	if(token.type != GrammarFromStringLexer::TokenType::COMMA) {
		throw exception::CommonException("Unrecognised Comma token.");
	}

	std::map<std::tuple<std::vector<DefaultSymbolType>, DefaultSymbolType, std::vector<DefaultSymbolType>>, std::set<std::vector<DefaultSymbolType>>> rules = parsePreservingCSLikeRules(input);

	token = GrammarFromStringLexer::next(input);
	if(token.type != GrammarFromStringLexer::TokenType::COMMA) {
		throw exception::CommonException("Unrecognised Comma token.");
	}

	DefaultSymbolType initialSymbol = alib::stringApi<DefaultSymbolType>::parse(input);

	token = GrammarFromStringLexer::next(input);
	if(token.type != GrammarFromStringLexer::TokenType::TUPLE_END) {
		throw exception::CommonException("Unrecognised Tuple end token.");
	}

	T grammar(nonterminals, terminals, initialSymbol);
	for(const auto& rule : rules) {
		for(const auto& ruleRHS : rule.second) {
			grammar.addRule(std::get<0>(rule.first), std::get<1>(rule.first), std::get<2>(rule.first), ruleRHS);
		}
	}
	return grammar;
}

RightRG < > GrammarFromStringParser::parseRightRG(std::istream& input) {
	GrammarFromStringLexer::Token token = GrammarFromStringLexer::next(input);
	if(token.type != GrammarFromStringLexer::TokenType::RIGHT_RG) {
		throw exception::CommonException("Unrecognised RightRG token.");
	}

	return parseCFLikeGrammar<RightRG < > >(input);
}

LeftRG < > GrammarFromStringParser::parseLeftRG(std::istream& input) {
	GrammarFromStringLexer::Token token = GrammarFromStringLexer::next(input);
	if(token.type != GrammarFromStringLexer::TokenType::LEFT_RG) {
		throw exception::CommonException("Unrecognised LeftRG token.");
	}

	return parseCFLikeGrammar<LeftRG < > >(input);
}

RightLG < > GrammarFromStringParser::parseRightLG(std::istream& input) {
	GrammarFromStringLexer::Token token = GrammarFromStringLexer::next(input);
	if(token.type != GrammarFromStringLexer::TokenType::RIGHT_LG) {
		throw exception::CommonException("Unrecognised RightLG token.");
	}

	return parseCFLikeGrammar<RightLG < > >(input);
}

LeftLG < > GrammarFromStringParser::parseLeftLG(std::istream& input) {
	GrammarFromStringLexer::Token token = GrammarFromStringLexer::next(input);
	if(token.type != GrammarFromStringLexer::TokenType::LEFT_LG) {
		throw exception::CommonException("Unrecognised LeftLG token.");
	}

	return parseCFLikeGrammar<LeftLG < > >(input);
}

LG < > GrammarFromStringParser::parseLG(std::istream& input) {
	GrammarFromStringLexer::Token token = GrammarFromStringLexer::next(input);
	if(token.type != GrammarFromStringLexer::TokenType::LG) {
		throw exception::CommonException("Unrecognised LG token.");
	}

	return parseCFLikeGrammar<LG < > >(input);
}

CFG < > GrammarFromStringParser::parseCFG(std::istream& input) {
	GrammarFromStringLexer::Token token = GrammarFromStringLexer::next(input);
	if(token.type != GrammarFromStringLexer::TokenType::CFG) {
		throw exception::CommonException("Unrecognised CFG token.");
	}

	return parseCFLikeGrammar<CFG < > >(input);
}

EpsilonFreeCFG < > GrammarFromStringParser::parseEpsilonFreeCFG(std::istream& input) {
	GrammarFromStringLexer::Token token = GrammarFromStringLexer::next(input);
	if(token.type != GrammarFromStringLexer::TokenType::EPSILON_FREE_CFG) {
		throw exception::CommonException("Unrecognised EpsilonFreeCFG token.");
	}

	return parseCFLikeGrammar<EpsilonFreeCFG < > >(input);
}

GNF < > GrammarFromStringParser::parseGNF(std::istream& input) {
	GrammarFromStringLexer::Token token = GrammarFromStringLexer::next(input);
	if(token.type != GrammarFromStringLexer::TokenType::GNF) {
		throw exception::CommonException("Unrecognised GNF token.");
	}

	return parseCFLikeGrammar<GNF < > >(input);
}

CNF < > GrammarFromStringParser::parseCNF(std::istream& input) {
	GrammarFromStringLexer::Token token = GrammarFromStringLexer::next(input);
	if(token.type != GrammarFromStringLexer::TokenType::CNF) {
		throw exception::CommonException("Unrecognised CNF token.");
	}

	return parseCFLikeGrammar<CNF < > >(input);
}

NonContractingGrammar < > GrammarFromStringParser::parseNonContractingGrammar(std::istream& input) {
	GrammarFromStringLexer::Token token = GrammarFromStringLexer::next(input);
	if(token.type != GrammarFromStringLexer::TokenType::NON_CONTRACTING_GRAMMAR) {
		throw exception::CommonException("Unrecognised NonContractingGrammar token.");
	}

	return parseCSLikeGrammar<NonContractingGrammar < > >(input);
}

CSG < > GrammarFromStringParser::parseCSG(std::istream& input) {
	GrammarFromStringLexer::Token token = GrammarFromStringLexer::next(input);
	if(token.type != GrammarFromStringLexer::TokenType::CSG) {
		throw exception::CommonException("Unrecognised CSG token.");
	}

	return parsePreservingCSLikeGrammar< CSG < > >(input);
}

ContextPreservingUnrestrictedGrammar < > GrammarFromStringParser::parseContextPreservingUnrestrictedGrammar(std::istream& input) {
	GrammarFromStringLexer::Token token = GrammarFromStringLexer::next(input);
	if(token.type != GrammarFromStringLexer::TokenType::CONTEXT_PRESERVING_UNRESTRICTED_GRAMMAR) {
		throw exception::CommonException("Unrecognised ContextPreservingUnrestrictedGrammar token.");
	}

	return parsePreservingCSLikeGrammar<ContextPreservingUnrestrictedGrammar < > >(input);
}

UnrestrictedGrammar < > GrammarFromStringParser::parseUnrestrictedGrammar(std::istream& input) {
	GrammarFromStringLexer::Token token = GrammarFromStringLexer::next(input);
	if(token.type != GrammarFromStringLexer::TokenType::UNRESTRICTED_GRAMMAR) {
		throw exception::CommonException("Unrecognised UnrestrictedGrammar token.");
	}

	return parseCSLikeGrammar<UnrestrictedGrammar < > >(input);
}

} /* namespace grammar */
